package sun.security.krb5;

import sun.security.krb5.internal.Krb5;
import sun.security.krb5.internal.crypto.EType;
import sun.security.krb5.internal.KrbApErrException;
import java.util.Arrays;
import sun.security.krb5.internal.HostAddress;
import sun.security.krb5.internal.rcache.AuthTimeWithHash;
import java.security.NoSuchAlgorithmException;
import java.security.MessageDigest;
import sun.security.krb5.internal.EncTicketPart;
import sun.security.krb5.internal.KRBError;
import sun.security.util.DerValue;
import sun.security.krb5.internal.KdcErrException;
import sun.security.krb5.internal.AuthorizationData;
import sun.security.krb5.internal.SeqNumber;
import sun.security.krb5.internal.Ticket;
import java.net.InetAddress;
import sun.security.jgss.krb5.Krb5AcceptCredential;
import java.io.IOException;
import sun.security.krb5.internal.LocalSeqNumber;
import sun.security.krb5.internal.APOptions;
import sun.security.krb5.internal.ReplayCache;
import sun.security.krb5.internal.APReq;
import sun.security.krb5.internal.Authenticator;
import sun.security.krb5.internal.KerberosTime;

public class KrbApReq
{
    private byte[] obuf;
    private KerberosTime ctime;
    private int cusec;
    private Authenticator authenticator;
    private Credentials creds;
    private APReq apReqMessg;
    private static ReplayCache rcache;
    private static boolean DEBUG;
    private static final char[] hexConst;
    
    public KrbApReq(Credentials credentials, boolean b, boolean b2, boolean b3, Checksum checksum) throws Asn1Exception, KrbCryptoException, KrbException, IOException {
        APOptions obj = b ? new APOptions(2) : new APOptions();
        if (KrbApReq.DEBUG) {
            System.out.println(">>> KrbApReq: APOptions are " + obj);
        }
        this.init(obj, credentials, checksum, b2 ? new EncryptionKey(credentials.getSessionKey()) : null, new LocalSeqNumber(), null, 11);
    }
    
    public KrbApReq(byte[] obuf, Krb5AcceptCredential krb5AcceptCredential, InetAddress inetAddress) throws KrbException, IOException {
        this.obuf = obuf;
        if (this.apReqMessg == null) {
            this.decode();
        }
        this.authenticate(krb5AcceptCredential, inetAddress);
    }
    
    KrbApReq(APOptions apOptions, Ticket ticket, EncryptionKey encryptionKey, PrincipalName principalName, Checksum checksum, KerberosTime kerberosTime, EncryptionKey encryptionKey2, SeqNumber seqNumber, AuthorizationData authorizationData) throws Asn1Exception, IOException, KdcErrException, KrbCryptoException {
        this.init(apOptions, ticket, encryptionKey, principalName, checksum, kerberosTime, encryptionKey2, seqNumber, authorizationData, 7);
    }
    
    private void init(APOptions apOptions, Credentials credentials, Checksum checksum, EncryptionKey encryptionKey, SeqNumber seqNumber, AuthorizationData authorizationData, int n) throws KrbException, IOException {
        this.ctime = KerberosTime.now();
        this.init(apOptions, credentials.ticket, credentials.key, credentials.client, checksum, this.ctime, encryptionKey, seqNumber, authorizationData, n);
    }
    
    private void init(APOptions apOptions, Ticket ticket, EncryptionKey encryptionKey, PrincipalName principalName, Checksum checksum, KerberosTime kerberosTime, EncryptionKey encryptionKey2, SeqNumber seqNumber, AuthorizationData authorizationData, int n) throws Asn1Exception, IOException, KdcErrException, KrbCryptoException {
        this.createMessage(apOptions, ticket, encryptionKey, principalName, checksum, kerberosTime, encryptionKey2, seqNumber, authorizationData, n);
        this.obuf = this.apReqMessg.asn1Encode();
    }
    
    void decode() throws KrbException, IOException {
        this.decode(new DerValue(this.obuf));
    }
    
    void decode(DerValue derValue) throws KrbException, IOException {
        this.apReqMessg = null;
        try {
            this.apReqMessg = new APReq(derValue);
        }
        catch (Asn1Exception cause) {
            this.apReqMessg = null;
            KRBError krbError = new KRBError(derValue);
            String errorString = krbError.getErrorString();
            String substring;
            if (errorString.charAt(errorString.length() - 1) == '\0') {
                substring = errorString.substring(0, errorString.length() - 1);
            }
            else {
                substring = errorString;
            }
            KrbException o = new KrbException(krbError.getErrorCode(), substring);
            o.initCause(cause);
            throw o;
        }
    }
    
    private void authenticate(Krb5AcceptCredential krb5AcceptCredential, InetAddress inetAddress) throws KrbException, IOException {
        int eType = this.apReqMessg.ticket.encPart.getEType();
        EncryptionKey key = EncryptionKey.findKey(eType, this.apReqMessg.ticket.encPart.getKeyVersionNumber(), krb5AcceptCredential.getKrb5EncryptionKeys(this.apReqMessg.ticket.sname));
        if (key == null) {
            throw new KrbException(400, "Cannot find key of appropriate type to decrypt AP REP - " + EType.toString(eType));
        }
        EncTicketPart encTicketPart = new EncTicketPart(this.apReqMessg.ticket.encPart.reset(this.apReqMessg.ticket.encPart.decrypt(key, 2)));
        checkPermittedEType(encTicketPart.key.getEType());
        this.authenticator = new Authenticator(this.apReqMessg.authenticator.reset(this.apReqMessg.authenticator.decrypt(encTicketPart.key, 11)));
        this.ctime = this.authenticator.ctime;
        this.cusec = this.authenticator.cusec;
        this.authenticator.ctime = this.authenticator.ctime.withMicroSeconds(this.authenticator.cusec);
        if (!this.authenticator.cname.equals(encTicketPart.cname)) {
            throw new KrbApErrException(36);
        }
        if (!this.authenticator.ctime.inClockSkew()) {
            throw new KrbApErrException(37);
        }
        byte[] digest;
        try {
            digest = MessageDigest.getInstance("MD5").digest(this.apReqMessg.authenticator.cipher);
        }
        catch (NoSuchAlgorithmException ex) {
            throw new AssertionError("Impossible");
        }
        char[] value = new char[digest.length * 2];
        for (int i = 0; i < digest.length; ++i) {
            value[2 * i] = KrbApReq.hexConst[(digest[i] & 0xFF) >> 4];
            value[2 * i + 1] = KrbApReq.hexConst[digest[i] & 0xF];
        }
        KrbApReq.rcache.checkAndStore(KerberosTime.now(), new AuthTimeWithHash(this.authenticator.cname.toString(), this.apReqMessg.ticket.sname.toString(), this.authenticator.ctime.getSeconds(), this.authenticator.cusec, new String(value)));
        if (inetAddress != null) {
            HostAddress addr = new HostAddress(inetAddress);
            if (encTicketPart.caddr != null && !encTicketPart.caddr.inList(addr)) {
                if (KrbApReq.DEBUG) {
                    System.out.println(">>> KrbApReq: initiator is " + addr.getInetAddress() + ", but caddr is " + Arrays.toString(encTicketPart.caddr.getInetAddresses()));
                }
                throw new KrbApErrException(38);
            }
        }
        KerberosTime now = KerberosTime.now();
        if ((encTicketPart.starttime != null && encTicketPart.starttime.greaterThanWRTClockSkew(now)) || encTicketPart.flags.get(7)) {
            throw new KrbApErrException(33);
        }
        if (encTicketPart.endtime != null && now.greaterThanWRTClockSkew(encTicketPart.endtime)) {
            throw new KrbApErrException(32);
        }
        this.creds = new Credentials(this.apReqMessg.ticket, this.authenticator.cname, null, this.apReqMessg.ticket.sname, null, encTicketPart.key, encTicketPart.flags, encTicketPart.authtime, encTicketPart.starttime, encTicketPart.endtime, encTicketPart.renewTill, encTicketPart.caddr, encTicketPart.authorizationData);
        if (KrbApReq.DEBUG) {
            System.out.println(">>> KrbApReq: authenticate succeed.");
        }
        return;
    }
    
    public Credentials getCreds() {
        return this.creds;
    }
    
    KerberosTime getCtime() {
        if (this.ctime != null) {
            return this.ctime;
        }
        return this.authenticator.ctime;
    }
    
    int cusec() {
        return this.cusec;
    }
    
    APOptions getAPOptions() throws KrbException, IOException {
        if (this.apReqMessg == null) {
            this.decode();
        }
        if (this.apReqMessg != null) {
            return this.apReqMessg.apOptions;
        }
        return null;
    }
    
    public boolean getMutualAuthRequired() throws KrbException, IOException {
        if (this.apReqMessg == null) {
            this.decode();
        }
        return this.apReqMessg != null && this.apReqMessg.apOptions.get(2);
    }
    
    boolean useSessionKey() throws KrbException, IOException {
        if (this.apReqMessg == null) {
            this.decode();
        }
        return this.apReqMessg != null && this.apReqMessg.apOptions.get(1);
    }
    
    public EncryptionKey getSubKey() {
        return this.authenticator.getSubKey();
    }
    
    public Integer getSeqNumber() {
        return this.authenticator.getSeqNumber();
    }
    
    public Checksum getChecksum() {
        return this.authenticator.getChecksum();
    }
    
    public byte[] getMessage() {
        return this.obuf;
    }
    
    public PrincipalName getClient() {
        return this.creds.getClient();
    }
    
    private void createMessage(APOptions new_apOptions, Ticket new_ticket, EncryptionKey encryptionKey, PrincipalName new_cname, Checksum new_cksum, KerberosTime new_ctime, EncryptionKey new_subKey, SeqNumber seqNumber, AuthorizationData new_authorizationData, int n) throws Asn1Exception, IOException, KdcErrException, KrbCryptoException {
        Integer new_seqNumber = null;
        if (seqNumber != null) {
            new_seqNumber = new Integer(seqNumber.current());
        }
        this.authenticator = new Authenticator(new_cname, new_cksum, new_ctime.getMicroSeconds(), new_ctime, new_subKey, new_seqNumber, new_authorizationData);
        this.apReqMessg = new APReq(new_apOptions, new_ticket, new EncryptedData(encryptionKey, this.authenticator.asn1Encode(), n));
    }
    
    private static void checkPermittedEType(int n) throws KrbException {
        if (!EType.isSupported(n, EType.getDefaults("permitted_enctypes"))) {
            throw new KrbException(EType.toString(n) + " encryption type not in permitted_enctypes list");
        }
        return;
    }
    
    static {
        KrbApReq.rcache = ReplayCache.getInstance();
        KrbApReq.DEBUG = Krb5.DEBUG;
        hexConst = "0123456789ABCDEF".toCharArray();
    }
}
